import { Dispatch, Store } from 'redux';
import { ClassConstructor } from '.';
import { FitDecorated, TemplateFunction } from './fit-element';
/**
 * The function that extracts required data out of the state and the passed props.
 */
export declare type MapStateToPropsFn<S, P, OP> = (state: S, ownProps: OP) => P;
/**
 * The function that sets up the view actions based on the store's dispatch function.
 */
export declare type MapDispatchToPropsFn<S, P, OP> = (dispatch: Dispatch<S>, ownProps: OP) => P;
/**
 * A factory method for creating a specialized {@ref MapStateToPropsFn<S, P, OP>}
 * for every instance of the component.
 *
 * Will be invoked at component construction time.
 */
export declare type MapStateToPropsFactory<S, P, OP> = () => MapStateToPropsFn<S, P, OP>;
/**
 * An element connected to the redux store.
 */
export interface ConnectElement<S> {
    getStore(): Store<S>;
}
export declare type ConnectedElement<S, SP, DP, OP, B extends ClassConstructor<HTMLElement>, T extends FitDecorated<B, OP, SP & DP>> = T & ClassConstructor<ConnectElement<S>>;
/**
 * The type of decorator wrapping a templating function.
 */
export declare type TemplateDecorator<S, SP, DP, OP> = (fn: TemplateFunction<SP & DP>) => ConnectedElement<S, SP, DP, OP, typeof HTMLElement, FitDecorated<typeof HTMLElement, OP, SP & DP>>;
/**
 * The type of decorator wrapping a connected element.
 */
export declare type ElementDecorator<S, SP, DP, OP> = <B extends ClassConstructor<HTMLElement>, T extends FitDecorated<B, OP, SP & DP>>(base: T) => ConnectedElement<S, SP, DP, OP, B, T>;
/**
 * Creates a decorator to create 💪 web components using the given template function / base class.
 *
 * @param {MapStateToPropsFn<S, SP, OP>} mapStateToProps The MapStateToProps function or factory.
 * @param {MapDispatchToPropsFn<S, DP, OP>} mapDispatchToProps The MapStateToDispatch function.
 * @returns The decorator function.
 * @template S, SP, DP, OP
 */
export default function connect<S, SP, DP, OP = {}>(mapStateToProps: MapStateToPropsFactory<S, SP, OP> | MapStateToPropsFn<S, SP, OP>, mapDispatchToProps: MapDispatchToPropsFn<S, DP, OP> | DP): TemplateDecorator<S, SP, DP, OP> & ElementDecorator<S, SP, DP, OP>;
