import { TemplateResult } from 'lit-html';
import { ClassConstructor } from '.';
/**
 * A lit-html template construction function.
 */
export declare type TemplateFunction<P> = (props: P) => TemplateResult;
/**
 * A value constructor.
 */
export declare type Transformer = (value: string) => any;
/**
 * Attribute observer configuration.
 *
 * If a constructor function is present, it will be invoked when an
 * attribute changes with the attribute's new value. If null is passed
 * or a property is set the value will be left untouched.
 */
export declare type AttributeDescriptors<OP> = {
    [key in keyof OP]: Transformer | null;
};
/**
 * A custom element with its lifecycle callbacks.
 */
export interface CustomElement extends HTMLElement {
    attributeChangedCallback?(name: string, oldValue: string, newValue: string): any;
    connectedCallback?(): any;
    disconnectedCallback?(): any;
}
/**
 * A constructor of a custom element.
 */
export interface CustomElementConstructor {
    observedAttributes?: string[];
    new (...args: any[]): CustomElement;
}
/**
 * A 💪-html decorated element.
 */
export interface FitElement<OP, RP> {
    ownProps: OP;
    renderProps: RP;
    template: TemplateFunction<RP>;
    attributeChangedCallback(name: string, oldValue: string, newValue: string): any;
    connectedCallback(): any;
    disconnectedCallback(): any;
    enqueueRender(): any;
    render(): any;
}
/**
 * The constructor for a 💪-html element.
 */
export interface FitElementConstructor<OP, RP> {
    observedAttributes: string[];
    properties: AttributeDescriptors<OP>;
    new (...args: any[]): FitElement<OP, RP>;
}
/**
 * A 💪-html decorated class.
 */
export declare type FitDecorated<B extends ClassConstructor<HTMLElement>, OP, RP> = B & FitElementConstructor<OP, RP>;
/**
 * Creates a subclass of the given HTML element that uses lit-html rendering and listens
 * for attribute and property changes.
 *
 * @param {TemplateFunction<RP>} templ The lit-html templating function.
 * @param {AttributeDescriptors<OP>} desc Property descriptors to enable the attribute & property listening.
 * @returns The decorator function.
 * @template OP, RP
 */
export default function withFit<OP, RP = OP>(templ: TemplateFunction<RP>, desc?: AttributeDescriptors<OP>): <T extends CustomElementConstructor>(base: T) => FitDecorated<T, OP, RP>;
