import { render as shadyRender } from 'lit-html/lib/shady-render';
import { isFunction, shallowEqual } from './util';
/**
 * Creates a subclass of the given HTML element that uses lit-html rendering and listens
 * for attribute and property changes.
 *
 * @param {TemplateFunction<RP>} templ The lit-html templating function.
 * @param {AttributeDescriptors<OP>} desc Property descriptors to enable the attribute & property listening.
 * @returns The decorator function.
 * @template OP, RP
 */
export default function withFit(templ, desc) {
    return (base) => {
        const Element = class extends base {
            constructor(...args) {
                super(...args);
                this._isConnected = false;
                this._nodeName = this.nodeName.toLowerCase();
                this._renderEnqueued = false;
                // Only iterate over properties declared in _this_ class, not over
                // the ones from the superclass.
                for (const propName of Object.keys(Element.properties)) {
                    Object.defineProperty(this, propName, {
                        configurable: true,
                        enumerable: true,
                        get: () => this.ownProps
                            ? this.ownProps[propName]
                            : undefined,
                        set: val => {
                            if (this[propName] === val) {
                                return;
                            }
                            this.ownProps = Object.assign({}, this.ownProps, { [propName]: val });
                        },
                    });
                }
                this.attachShadow({ mode: 'open' });
            }
            static get observedAttributes() {
                const propKeys = Object.keys(this.properties);
                return (base.observedAttributes || [])
                    .filter(attr => propKeys.indexOf(attr) === -1)
                    .concat(propKeys);
            }
            static get properties() {
                return desc || {};
            }
            get ownProps() {
                return this.renderProps;
            }
            set ownProps(props) {
                this.renderProps = props;
            }
            get renderProps() {
                return this._renderProps;
            }
            set renderProps(props) {
                if (shallowEqual(props, this._renderProps)) {
                    return;
                }
                this._renderProps = props;
                this.enqueueRender();
            }
            get template() {
                return templ;
            }
            attributeChangedCallback(name, oldValue, newValue) {
                if (isFunction(super.attributeChangedCallback)) {
                    super.attributeChangedCallback(name, oldValue, newValue);
                }
                if (!(name in Element.properties) || oldValue === newValue) {
                    return;
                }
                const transformer = Element.properties[name];
                if (!transformer) {
                    this[name] = newValue;
                }
                else if (transformer === Boolean) {
                    // tslint:disable-next-line:triple-equals
                    this[name] = newValue != null;
                }
                else {
                    this[name] = transformer(newValue);
                }
            }
            connectedCallback() {
                if (isFunction(super.connectedCallback)) {
                    super.connectedCallback();
                }
                if (window.ShadyCSS) {
                    window.ShadyCSS.styleElement(this);
                }
                this._isConnected = true;
            }
            disconnectedCallback() {
                if (isFunction(super.disconnectedCallback)) {
                    super.disconnectedCallback();
                }
                this._isConnected = false;
            }
            enqueueRender() {
                if (this._renderEnqueued) {
                    return;
                }
                this._renderEnqueued = true;
                Promise.resolve().then(() => {
                    this._renderEnqueued = false;
                    if (this._isConnected) {
                        this.render();
                    }
                });
            }
            render() {
                shadyRender(this.template(this.renderProps), this.shadowRoot, this._nodeName);
            }
        };
        return Element;
    };
}
//# sourceMappingURL=fit-element.js.map